// Doom to Delphi Total Conversion
// Author: Jim Valavanis
// E-Mail: jimmyvalavanis@yahoo.gr
// Site  : http://www.geocities.com/jimmyvalavanis/applications/delphidoom.html

unit d_player;

interface

uses
// The player data structure depends on a number
// of other structs: items (internal inventory),
// animation states (closely tied to the sprites
// used to represent them, unfortunately).
  d_items, p_pspr_h,
// In addition, the player is just a special
// case of the generic moving object/actor.
  p_mobj_h,
// Finally, for odd reasons, the player input
// is buffered within the player data struct,
// as commands per game tick.
  d_ticcmd,
  m_fixed, doomdef;

{
    d_player.h
}

// Emacs style mode select   -*- C++ -*-  
//----------------------------------------------------------------------------- 
// 
// $Id:$ 
// 
// Copyright (C) 1993-1996 by id Software, Inc. 
// 
// This source is available for distribution and/or modification 
// only under the terms of the DOOM Source Code License as 
// published by id Software. All rights reserved. 
// 
// The source is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License 
// for more details. 
// 
// DESCRIPTION: 
// 
// 
//----------------------------------------------------------------------------- 

// 
// Player states. 
//

type
  playerstate_t = (
  { Playing or camping. }
    PST_LIVE,
  { Dead on the ground, view follows killer. }
    PST_DEAD,
  { Ready to restart/respawn??? }
    PST_REBORN);

  { }
  { Player internal flags, for cheats and debug. }
  { }

const
  { No clipping, walk through barriers. }
    CF_NOCLIP = 1;
  { No damage, no health loss. }
    CF_GODMODE = 2;
  { Not really a cheat, just a debug aid. }
    CF_NOMOMENTUM = 4;

type
  { }
  { Extended player object info: player_t }
  { }
  player_t = record
    mo : Pmobj_t;
    playerstate : playerstate_t;
    cmd : ticcmd_t;

    // Determine POV,
    //  including viewpoint bobbing during movement.
    // Focal origin above r.z
    viewz : fixed_t;
    // Base height above floor for viewz.
    viewheight : fixed_t;
    // Bob/squat speed.
    deltaviewheight: fixed_t;
    // bounded/scaled total momentum.
    bob: fixed_t;

    lookdir: integer;
    centering: boolean;

    // This is only used between levels,
    // mo->health is used during levels.
    health : integer;
    armorpoints : integer;
    // Armor type is 0-2.
    armortype : integer;

    // Power ups. invinc and invis are tic counters.
    powers: array[0..Ord(NUMPOWERS) - 1] of integer;
    cards: array[0..Ord(NUMCARDS) - 1] of boolean;
    backpack: boolean;

    // Frags, kills of other players.
    frags: array[0..(MAXPLAYERS)-1] of integer;
    readyweapon: weapontype_t;

    // Is wp_nochange if not changing.
    pendingweapon : weapontype_t;

    weaponowned: array[0..Ord(NUMWEAPONS) - 1] of integer;
    ammo: array[0..Ord(NUMAMMO) - 1] of integer;
    maxammo: array[0..Ord(NUMAMMO) - 1] of integer;

    // True if button down last tic.
    attackdown : boolean;
    usedown : boolean;

    // Bit flags, for cheats and debug.
    // See cheat_t, above.
    cheats : integer;

    // Refired shots are less accurate.
    refire : integer;

    // For intermission stats.
    killcount : integer;
    itemcount : integer;
    secretcount : integer;

    // Hint messages.
    _message : string[255];

    // For screen flashing (red or bright).
    damagecount : integer;
    bonuscount : integer;

    // Who did damage (NULL for floors/ceilings).
    attacker : Pmobj_t;

    // So gun flashes light up areas.
    extralight : integer;

    // Current PLAYPAL, ???
    //  can be set to REDCOLORMAP for pain, etc.
    fixedcolormap : integer;

    // Player skin colorshift,
    //  0-3 for which color to draw player.
    colormap : integer;

    // Overlay view sprites (gun, etc).
    psprites : array[0..Ord(NUMPSPRITES) - 1] of pspdef_t;

    // True if secret level has been done.
    didsecret : boolean;
  end;

  player_s = player_t;

  Pplayer_s = ^player_s;
  Pplayer_t = ^player_t;

  { }
  { INTERMISSION }
  { Structure passed e.g. to WI_Start(wb) }
  { }

  wbplayerstruct_t = record
    _in : boolean; // whether the player is in game

    // Player stats, kills, collected items etc.
    skills : integer;
    sitems : integer;
    ssecret : integer;
    stime : integer;
    frags : array[0..3] of integer;
    score : integer; // current score on entry, modified on return
  end;
  Pwbplayerstruct_t = ^wbplayerstruct_t;
  wbplayerstruct_tArray = packed array[0..$FFFF] of wbplayerstruct_t;
  Pwbplayerstruct_tArray = ^wbplayerstruct_tArray;

  wbstartstruct_t = record
    epsd: integer; // episode # (0-2)

    // if true, splash the secret level
    didsecret: boolean;

    // previous and next levels, origin 0
    last: integer;
    next: integer;

    maxkills: integer;
    maxitems: integer;
    maxsecret: integer;
    maxfrags: integer;

    // the par time
    partime: integer;

    // index of this player in game
    pnum: integer;

    plyr: array[0..MAXPLAYERS-1] of wbplayerstruct_t;
  end;
  Pwbstartstruct_t = ^wbstartstruct_t;

var
// VJ -> moved from g_game
  players: array[0..MAXPLAYERS - 1] of player_t;

implementation


end.

