// Doom to Delphi Total Conversion
// Author: Jim Valavanis
// E-Mail: jimmyvalavanis@yahoo.gr
// Site  : http://www.geocities.com/jimmyvalavanis/applications/delphidoom.html

unit f_wipe;

interface

{
    f_wipe.h, f_wipe.c
}

// Emacs style mode select   -*- C++ -*-  
//----------------------------------------------------------------------------- 
// 
// $Id:$ 
// 
// Copyright (C) 1993-1996 by id Software, Inc. 
// 
// This source is available for distribution and/or modification 
// only under the terms of the DOOM Source Code License as 
// published by id Software. All rights reserved. 
// 
// The source is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License 
// for more details. 
// 
// DESCRIPTION: 
//  Mission start screen wipe/melt, special effects. 
// Mission begin melt/wipe screen special effect. 
//   
//----------------------------------------------------------------------------- 

function wipe_StartScreen(x, y, width, height: integer): integer;

function wipe_EndScreen(x, y, width, height: integer): integer;

function wipe_ScreenWipe(wipeno, x, y, width, height, ticks: integer): boolean;

type
  wipe_t = (
    // simple gradual pixel change for 8-bit only
    wipe_ColorXForm,
    // weird screen melt
    wipe_Melt,
    wipe_NUMWIPES
  );

implementation

uses doomdef, d_delphi,
  m_rnd, m_fixed,
  i_video, v_video,
  z_zone;

//
//                       SCREEN WIPE PACKAGE
//

// when zero, stop the wipe
var
  go: boolean = false;

var  
  wipe_scr_start: PByteArray;
  wipe_scr_end: PByteArray;
  wipe_scr: PByteArray;

procedure wipe_shittyColMajorXform(_array: PByteArray; width, height: integer);
var
  x, y: integer;
  dest: PByteArray;
begin
  dest := PByteArray(Z_Malloc(width * height, PU_STATIC, nil));

  for y := 0 to height - 1 do
    for x := 0 to width - 1 do
      dest[x * height + y] := _array[y * width + x];

  memcpy(_array, dest, width * height);

  Z_Free(dest);
end;

function wipe_initColorXForm(width, height, ticks: integer): integer;
begin
  memcpy(wipe_scr, wipe_scr_start, width * height);
  result := 0;
end;

function wipe_doColorXForm(width, height, ticks: integer): integer;
var
  changed: boolean;
  w: integer;
  e: integer;
  newval: integer;
begin
  changed := false;
  w := 0;
  e := 0;

  while w < width * height do
  begin
    if wipe_scr[w] <> wipe_scr_end[e] then
    begin
      if wipe_scr[w] > wipe_scr_end[e] then
      begin
        newval := wipe_scr[w] - ticks;
        if newval < wipe_scr_end[e] then
          wipe_scr[w] := wipe_scr_end[e]
        else
          wipe_scr[w] := newval;
      end
      else if wipe_scr[w] < wipe_scr_end[e] then
      begin
        newval := wipe_scr[w] + ticks;
        if newval > wipe_scr_end[e] then
          wipe_scr[w] := wipe_scr_end[e]
        else
          wipe_scr[w] := newval;
      end;
      changed := true;
    end;
    inc(w);
    inc(e);
  end;

  result := intval(not changed);
end;

function wipe_exitColorXForm(width, height, ticks: integer): integer;
begin
  result := 0;
end;

var
  y: Pfixed_tArray;
  vy: fixed_t;

function wipe_initMelt(width, height, ticks: integer): integer;
var
  i, r: integer;
begin
  // copy start screen to main screen
  memcpy(wipe_scr, wipe_scr_start, width * height);

  wipe_shittyColMajorXform(wipe_scr_start, width, height);
  wipe_shittyColMajorXform(wipe_scr_end, width, height);

  // setup initial column positions
  // (y<0 => not ready to scroll yet)
  y := Pfixed_tArray(Z_Malloc(width * SizeOf(integer), PU_STATIC, nil));
  y[0] := -(M_Random mod 16);
  for i := 1 to width - 1 do
  begin
    r := (M_Random mod 3) - 1;
    y[i] := y[i - 1] + r;
    if y[i] > 0 then
      y[i] := 0
    else if y[i] = -16 then
      y[i] := -15;
  end;

  // VJ change wipe timing
  vy := FRACUNIT * SCREENWIDTH div 200;
  for i := 0 to width - 1 do
    y[i] := y[i] * vy;

  result := 0;
end;

function wipe_doMelt(width, height, ticks: integer): integer;
var
  i: integer;
  j: integer;
  dy: fixed_t;
  idx: integer;
  s: PByteArray;
  d: PByteArray;
begin
  result := 1;

  while ticks > 0 do
  begin
    for i := 0 to width - 1 do
    begin
      if y[i] < 0 then
      begin
        y[i] := y[i] + vy;
        result := 0;
      end
      else if y[i] < height * FRACUNIT then
      begin
        if y[i] <= 15 * vy then
          dy := y[i] + vy
        else
          dy := 8 * vy;
        if (y[i] + dy) div FRACUNIT >= height then
          dy := height * FRACUNIT - y[i];
        s := PByteArray(integer(wipe_scr_end) + i * height + y[i] div FRACUNIT);
        d := PByteArray(integer(wipe_scr) + y[i] div FRACUNIT * width + i);
        idx := 0;
        for j := 0 to dy div FRACUNIT do //- 1 do
        begin
          d[idx] := s[j];
          idx := idx + width;
        end;
        y[i] := y[i] + dy;
        s := PByteArray(integer(wipe_scr_start) + i * height);
        d := PByteArray(integer(wipe_scr) + y[i] div FRACUNIT * width + i);

        idx := 0;
        for j := 0 to height - y[i] div FRACUNIT - 1 do
        begin
          d[idx] := s[j];
          idx := idx + width;
        end;
        result := 0;
      end;
    end;
    dec(ticks);
  end;
end;

function wipe_exitMelt(width, height, ticks: integer): integer;
begin
  Z_Free(y);
  result := 0;
end;

function wipe_StartScreen(x, y, width, height: integer): integer;
begin
  wipe_scr_start := screens[2];
  I_ReadScreen(wipe_scr_start);
  result := 0;
end;

function wipe_EndScreen(x, y, width, height: integer): integer;
begin
  wipe_scr_end := screens[3];
  I_ReadScreen(wipe_scr_end);
  V_DrawBlock(x, y, 0, width, height, wipe_scr_start); // restore start scr.
  result := 0;
end;

function wipe_ScreenWipe(wipeno, x, y, width, height, ticks: integer): boolean;
var
  rc: integer;

  function WIPES(index: integer): integer;
  begin
    case index of
      0: result := wipe_initColorXForm(width, height, ticks);
      1: result := wipe_doColorXForm(width, height, ticks);
      2: result := wipe_exitColorXForm(width, height, ticks);
      3: result := wipe_initMelt(width, height, ticks);
      4: result := wipe_doMelt(width, height, ticks);
      5: result := wipe_exitMelt(width, height, ticks);
    else
      result := 0; // Ouch
    end;
  end;

begin
  // initial stuff
  if not go then
  begin
    go := true;
  // wipe_scr = (byte *) Z_Malloc(width*height, PU_STATIC, 0); // DEBUG
    wipe_scr := screens[_FG];

    WIPES(wipeno * 3)
  end;

  // do a piece of wipe-in
  V_MarkRect(0, 0, width, height, false);
  rc := WIPES(wipeno * 3 + 1);
    //  V_DrawBlock(x, y, 0, width, height, wipe_scr); // DEBUG

  // final stuff
  if rc <> 0 then
  begin
    go := false;
    WIPES(wipeno * 3 + 2);
  end;

  result := not go;
end;

end.

