unit i_net;

interface

{
    i_net.h, i_net.c
}

  { Emacs style mode select   -*- C++ -*-  }
  {----------------------------------------------------------------------------- }
  { }
  { $Id:$ }
  { }
  { Copyright (C) 1993-1996 by id Software, Inc. }
  { }
  { This source is available for distribution and/or modification }
  { only under the terms of the DOOM Source Code License as }
  { published by id Software. All rights reserved. }
  { }
  { The source is distributed in the hope that it will be useful, }
  { but WITHOUT ANY WARRANTY; without even the implied warranty of }
  { FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License }
  { for more details. }
  { }
  { DESCRIPTION: }
  {	System specific network interface stuff. }
  { }
  {----------------------------------------------------------------------------- }

{ Called by D_DoomMain. }

procedure I_InitNetwork;

procedure I_NetCmd;

implementation

uses Windows, WinSock,
  doomtype, i_system, d_event, d_net, m_argv, d_player,
  doomstat, d_delphi;

const
  IPPORT_USERRESERVED = 5000;

type
  socklen_t = integer;

//
// NETWORKING
//
var
  DOOMPORT: word;
  mysocket: TSocket;
  sendaddress: array[0..MAXNETNODES - 1] of sockaddr_in;
  sendplayer: array[0..MAXNETNODES] of byte;

var
  netget: PProcedure;
  netsend: PProcedure;

function neterror: string; forward;

const
	PRE_CONNECT = 0;
	PRE_DISCONNECT = 1;
	PRE_ALLHERE = 2;
	PRE_CONACK = 3;
	PRE_ALLHEREACK = 4;
	PRE_GO = 5;

// Set PreGamePacket.fake to this so that the game rejects any pregame packets
// after it starts. This translates to NCMD_SETUP|NCMD_MULTI.
  PRE_FAKE = $30;

type
  machine_t = record
    address: LongWord;
    port: word;
    player: byte;
    pad: byte;
  end;

  PreGamePacket_t = record
    fake: byte;
    _message: byte;
    numnodes: byte;
    consolenum: byte;
    machines: array[0..MAXNETNODES - 1] of machine_t;
  end;

//
// UDPsocket
//
function UDPsocket: TSocket;
begin
	// allocate a socket
	result := socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if result = INVALID_SOCKET then
    I_Error('UDPsocket(): Can not create socket: %s', [neterror]);
end;

//
// BindToLocalPort
//
procedure BindToLocalPort(s: TSocket; port: word);
var
  v: integer;
  address: sockaddr_in;
begin
	memset(@address, 0, SizeOf(address));
	address.sin_family := AF_INET;
	address.sin_addr.s_addr := INADDR_ANY;
	address.sin_port := htons(port);

  v := bind(s, address, SizeOf(address));
  if v = SOCKET_ERROR then
    I_Error('BindToLocalPort(): %s', [neterror]);
end;

function FindNode(const address: sockaddr_in): integer;
var
  i: integer;
begin
  // find remote node number
  i := 0;
  while i < doomcom.numnodes do
  begin
		if (address.sin_addr.s_addr = sendaddress[i].sin_addr.s_addr) and
       (address.sin_port = sendaddress[i].sin_port) then
      break;
    inc(i);
  end;

	if i = doomcom.numnodes then
		// packet is not from one of the players (new game broadcast?)
		i := -1;

  result := i;
end;

//
// PacketSend
//
procedure PacketSend;
var
  c: integer;
begin
	//printf ("sending %i\n",gametic);
  c := sendto(mysocket, doomcom.data, doomcom.datalength,
        0, sendaddress[doomcom.remotenode], SizeOf(sendaddress[doomcom.remotenode]));

	//	if (c == -1)
	//			I_Error ("SendPacket error: %s",strerror(errno));
end;

const
  BACKUPTICS = 36;
  MAX_MSGLEN = BACKUPTICS * 10;

//
// PacketGet
//
procedure PacketGet;
var
  c: integer;
  fromlen: socklen_t;
  fromaddress: sockaddr_in;
  node: integer;
  err: integer;
begin
	fromlen := SizeOf(fromaddress);
	c := recvfrom(mysocket, doomcom.data, MAX_MSGLEN, 0, fromaddress, fromlen);
	node := FindNode(fromaddress);

	if c = SOCKET_ERROR then
  begin
    err := WSAGetLastError;

    if (err = WSAECONNRESET) and (node >= 0) then
    begin
      // The remote node aborted unexpectedly, so pretend it sent an exit packet
			printf('The connection from player %d was dropped' + #13#10,
				[sendplayer[node]]);

			doomcom.data[0] = $80;  // NCMD_EXIT
			c := 1;
    end
		else if err <> WSAEWOULDBLOCK then
    begin
      I_Error('PacketGet(): %s', [neterror]);
    end
    else
    begin
      doomcom.remotenode := -1; // no packet
      exit;
    end;
  end;

  doomcom.remotenode := node;
  doomcom.datalength := smallint(c);
end;

var
  fromaddress_pg: sockaddr_in;

function PreGet(buffer: pointer; bufferlen: integer; noabort: boolean): PSOCKADDR;
var
  fromlen: socklen_t;
  c: integer;
  err: integer;
begin
  fromlen := SizeIf(fromaddress);
  c := recvfrom(mysocket, buffer^, bufferlen, 0, fromaddress_pg, fromlen);

  if c = SOCKET_ERROR then
  begin
    err := WSAGetLastError;
    if (err = WSAEWOULDBLOCK) or (noabort and (err = WSAECONNRESET)) then
    begin
      result := nil; // no packet
      exit;
    end;
    I_Error ('PreGet(): %s', neterror);
  end;
	result := @fromaddress_pg;
end;

procedure PreSend(buffer: pointer; bufferlen: integer; _to: PSOCKADDR);
begin
  sendto(mysocket, buffer^, bufferlen, 0, _to^, SizeOf(_to^));
end;

procedure BuildAddress(address: PSOCKADDR; var name: string)
var
  PHostEnt: hostentry;  // host information entry
  port: word;
  portpart: string;
  isnamed: boolean;
  _pos: integer;
  i: integer;
begin
  isnamed := false;

  address.sin_family := AF_INET;

  _pos := Pos(':', name);
  if _pos > 0 then
  begin
    portpart := '';
    for i := _pos + 1 to Length(name) do
      portpart := portpart + name[i];
    port := atoi(portpart);
    if port = 0 then
    begin
      printf('Weird port: %s (using %d)' + #13#10, [portpart, DOOMPORT];
      port := DOOMPORT;
    end
  end
	else
    port := DOOMPORT;

  address.sin_port := htons(port);

  for i := 1 to Length(name) do
  begin
    if not (name[i] in ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '.') then
    begin
      isnamed := true;
      break;
    end;
  end;

  if not isnamed then
  begin
    address.sin_addr.s_addr := inet_addr(PChar(name));
    printf('Node number %d address %s' + #13#10, [doomcom.numnodes, name]);
  end
  else
  begin
    hostentry := gethostbyname(PChar(name));
		if hostentry = nil then
      I_Error('gethostbyname(): Could not find %s' + #13#10 + '%s',
        [name, neterror]);
    address.sin_addr.s_addr := LongWord(hostentry.h_addr_list);
    printf('Node number %d hostname %s' + #13#10,
      [doomcom.numnodes, StrPas(hostentry.h_name)]);
  end;
end;

procedure CloseNetwork;
begin
  if mysocket <> INVALID_SOCKET then
  begin
    closesocket(mysocket);
    mysocket := INVALID_SOCKET;
  end;
	WSACleanup;
end;

procedure StartNetwork(autoPort: boolean);
var
  trueval: LongWord;
  wsad: WSADATA;
begin
  trueval := 1
	if WSAStartup($0101, wsad) <> 0 then
    I_Error('StartNetwork(): Could not initialize Windows Sockets');

//  atterm (CloseNetwork);
  CloseNetwork;

  netsend := PacketSend;
  netget := PacketGet;
  netgame := true;
  multiplayer := true;

  // create communication socket
  mysocket := UDPsocket;
  if autoPort then
    BindToLocalPort(mysocket, 0)
  else
    BindToLocalPort(mysocket, DOOMPORT);
  ioctlsocket(mysocket, FIONBIO, trueval);
end;

procedure WaitForPlayers(i: integer);
begin
	if i := myargc then
    I_Error('WaitForPlayers(): Not enough parameters after -net');

  StartNetwork(false);

	// parse player number and host list
	doomcom.consoleplayer := SmallInt(Ord(myargv[i + 1][1]) - Ord('1'));
  printf('Console player number: %d' + #13#10, [doomcom.consoleplayer]);

	doomcom.numnodes := 1;		// this node for sure

	inc(i, 2);
  while (i <= myargc) and (myargv[i][1] <> '-') and (myargv[i][1] <> '+') do
  begin
    BuildAddress(@sendaddress[doomcom.numnodes], myargv[i]);
    inc(doomcom.numnodes);
    inc(i);
  end;

  printf('Total players: %d' + #13#10, [doomcom.numnodes]);

  doomcom.id := DOOMCOM_ID;
  doomcom.numplayers := doomcom.numnodes;
end;

procedure SendAbort;
var
  dis: array[0..1] of byte;
begin
  dis[0] := PRE_FAKE;
  dis[1] := PRE_DISCONNECT;

  dec(doomcom.numnodes);
  while doomcom.numnodes > 0 do
  begin
    PreSend(@dis, 2, @sendaddress[doomcom.numnodes]);
    PreSend(@dis, 2, @sendaddress[doomcom.numnodes]);
    PreSend(@dis, 2, @sendaddress[doomcom.numnodes]);
    PreSend(@dis, 2, @sendaddress[doomcom.numnodes]);
    dec(doomcom.numnodes);
  end;
end;



initialization
  DOOMPORT := (IPPORT_USERRESERVED + 29);
  mysocket := INVALID_SOCKET;

end.

