// Doom to Delphi Total Conversion
// Author: Jim Valavanis
// E-Mail: jimmyvalavanis@yahoo.gr
// Site  : http://www.geocities.com/jimmyvalavanis/applications/delphidoom.html

unit m_misc;

interface

{
    m_misc.h, m_misc.c
}

// Emacs style mode select   -*- C++ -*-  
//----------------------------------------------------------------------------- 
// 
// $Id:$ 
// 
// Copyright (C) 1993-1996 by id Software, Inc. 
// 
// This source is available for distribution and/or modification 
// only under the terms of the DOOM Source Code License as 
// published by id Software. All rights reserved. 
// 
// The source is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License 
// for more details. 
// 
// DESCRIPTION: 
// Main loop menu stuff. 
// Default Config File. 
// PCX Screenshots. 
// 
//     
//----------------------------------------------------------------------------- 

//
// MISC 
//

function M_WriteFile(const name: string; source: pointer; length: integer): boolean;

function M_ReadFile(const name: string; var buffer: Pointer): integer;

procedure M_ScreenShot;

procedure M_LoadDefaults;

procedure M_SaveDefaults;

implementation

uses classes,
  d_delphi,
  doomtype, doomdef,
  dstrings, d_englsh,
  d_main, d_player,
  g_game,
  hu_stuff,
  m_menu, m_argv,
  i_system, i_video,
  s_sound,
  v_video,
  w_wad, z_zone,
  doomstat;

function M_WriteFile(const name: string; source: pointer; length: integer): boolean;
var
  handle: file;
  count: integer;
begin
  assign(handle, name);
  {$I-}
  rewrite(handle, 1);
  {$I+}
  if IOResult <> 0 then
  begin
    result := false;
    exit;
  end;

  BlockWrite(handle, source^, length, count);
  close(handle);

  result := count > 0;
end;

function M_ReadFile(const name: string; var buffer: Pointer): integer;
var
  handle: file;
  count: integer;
begin
  assign(handle, name);
  {$I-}
  reset(handle, 1);
  {$I+}
  if IOResult <> 0 then
    I_Error('M_ReadFile(): Could not read file %s', [name]);

  result := FileSize(handle);
  buffer := Z_Malloc(result, PU_STATIC, nil);
  BlockRead(handle, buffer^, result, count);
  close(handle);

  if count < result then
    I_Error('M_ReadFile(): Could not read file %s', [name]);
end;

type
  ttype_t = (tString, tInteger, tBoolean);

  default_t = record
    name: string;
    location: pointer;
    defaultsvalue: string;
    defaultivalue: integer;
    defaultbvalue: boolean;
    _type: ttype_t;
  end;

const
  NUMDEFAULTS = 43;

  defaults: array[0..NUMDEFAULTS - 1] of default_t = (
    (name: 'mouse_sensitivity';
     location: @mouseSensitivity;
     defaultivalue: 5;
     _type: tInteger),

    (name: 'screenwidth';
     location: @SCREENWIDTH;
     defaultivalue: 640;
     _type: tInteger),
    (name: 'screenheight';
     location: @SCREENHEIGHT;
     defaultivalue: 480;
     _type: tInteger),
    (name: 'fullscreen';
     location: @fullscreen;
     defaultbvalue: true;
     _type: tBoolean),

    (name: 'sfx_volume';
     location: @snd_SfxVolume;
     defaultivalue: 8;
     _type: tInteger),
    (name: 'music_volume';
     location: @snd_MusicVolume;
     defaultivalue: 8;
     _type: tInteger),

    (name: 'show_messages';
     location: @showMessages;
     defaultivalue: 1;
     _type: tInteger),

    (name: 'key_right';
     location: @key_right;
     defaultivalue: KEY_RIGHTARROW;
     _type: tInteger),
    (name: 'key_left';
     location: @key_left;
     defaultivalue: KEY_LEFTARROW;
     _type: tInteger),
    (name: 'key_up';
     location: @key_up;
     defaultivalue: KEY_UPARROW;
     _type: tInteger),
    (name: 'key_down';
     location: @key_down;
     defaultivalue: KEY_DOWNARROW;
     _type: tInteger),
    (name: 'key_strafeleft';
     location: @key_strafeleft;
     defaultivalue: Ord(',');
     _type: tInteger),
    (name: 'key_straferight';
     location: @key_straferight;
     defaultivalue: Ord('.');
     _type: tInteger),
    (name: 'key_fire';
     location: @key_fire;
     defaultivalue: KEY_RCTRL;
     _type: tInteger),
    (name: 'key_use';
     location: @key_use;
     defaultivalue: Ord(' ');
     _type: tInteger),
    (name: 'key_strafe';
     location: @key_strafe;
     defaultivalue: KEY_RALT;
     _type: tInteger),
    (name: 'key_speed';
     location: @key_speed;
     defaultivalue: KEY_RSHIFT;
     _type: tInteger),
    (name: 'key_lookup';
     location: @key_lookup;
     defaultivalue: KEY_PAGEDOWN;
     _type: tInteger),
    (name: 'key_lookdown';
     location: @key_lookdown;
     defaultivalue: KEY_PAGEUP;
     _type: tInteger),
    (name: 'key_lookcenter';
     location: @key_lookcenter;
     defaultivalue: KEY_INS;
     _type: tInteger),


    (name: 'use_mouse';
     location: @usemouse;
     defaultbvalue: false;
     _type: tBoolean),
    (name: 'mouseb_fire';
     location: @mousebfire;
     defaultivalue: 0;
     _type: tInteger),
    (name: 'mouseb_strafe';
     location: @mousebstrafe;
     defaultivalue: 1;
     _type: tInteger),
    (name: 'mouseb_forward';
     location: @mousebforward;
     defaultivalue: 2;
     _type: tInteger),

    (name: 'use_joystick';
     location: @usejoystick;
     defaultbvalue: false;
     _type: tBoolean),
    (name: 'joyb_fire';
     location: @joybfire;
     defaultivalue: 0;
     _type: tInteger),
    (name: 'joyb_strafe';
     location: @joybstrafe;
     defaultivalue: 1;
     _type: tInteger),
    (name: 'joyb_use';
     location: @joybuse;
     defaultivalue: 3;
     _type: tInteger),
    (name: 'joyb_speed';
     location: @joybspeed;
     defaultivalue: 2;
     _type: tInteger),

    (name: 'screenblocks';
     location: @screenblocks;
     defaultivalue: 9;
     _type: tInteger),
    (name: 'detaillevel';
     location: @detailLevel;
     defaultivalue: 0;
     _type: tInteger),

    (name: 'snd_channels';
     location: @numChannels;
     defaultivalue: 3;
     _type: tInteger),

    (name: 'usegamma';
     location: @usegamma;
     defaultivalue: 0;
     _type: tInteger),

    (name: 'chatmacro0';
     location: @chat_macros[0];
     defaultsvalue: HUSTR_CHATMACRO0;
     _type: tString),
    (name: 'chatmacro1';
     location: @chat_macros[1];
     defaultsvalue: HUSTR_CHATMACRO1;
     _type: tString),
    (name: 'chatmacro2';
     location: @chat_macros[2];
     defaultsvalue: HUSTR_CHATMACRO2;
     _type: tString),
    (name: 'chatmacro3';
     location: @chat_macros[3];
     defaultsvalue: HUSTR_CHATMACRO3;
     _type: tString),
    (name: 'chatmacro4';
     location: @chat_macros[4];
     defaultsvalue: HUSTR_CHATMACRO4;
     _type: tString),
    (name: 'chatmacro5';
     location: @chat_macros[5];
     defaultsvalue: HUSTR_CHATMACRO5;
     _type: tString),
    (name: 'chatmacro6';
     location: @chat_macros[6];
     defaultsvalue: HUSTR_CHATMACRO6;
     _type: tString),
    (name: 'chatmacro7';
     location: @chat_macros[7];
     defaultsvalue: HUSTR_CHATMACRO7;
     _type: tString),
    (name: 'chatmacro8';
     location: @chat_macros[8];
     defaultsvalue: HUSTR_CHATMACRO8;
     _type: tString),
    (name: 'chatmacro9';
     location: @chat_macros[9];
     defaultsvalue: HUSTR_CHATMACRO9;
     _type: tString)
  );

var
  defaultfile: string;

procedure M_SaveDefaults;
var
  i: integer;
  s: TStringList;
begin
  s := TStringList.Create;
  try
    for i := 0 to NUMDEFAULTS - 1 do
      if defaults[i]._type = tInteger then
        s.Add(defaults[i].name + '=' + itoa(PInteger(defaults[i].location)^))
      else if defaults[i]._type = tString then
        s.Add(defaults[i].name + '=' + PString(defaults[i].location)^)
      else if defaults[i]._type = tBoolean then
        s.Add(defaults[i].name + '=' + itoa(IntVal(PBoolean(defaults[i].location)^)));

    s.SaveToFile(defaultfile);

  finally
    s.Free;
  end;
end;

procedure M_LoadDefaults;
var
  i: integer;
  j: integer;
  idx: integer;
  s: TStringList;
  n: string;
begin
  // set everything to base values
  for i := 0 to NUMDEFAULTS - 1 do
    if defaults[i]._type = tInteger then
      PInteger(defaults[i].location)^ := defaults[i].defaultivalue
    else if defaults[i]._type = tBoolean then
      PBoolean(defaults[i].location)^ := defaults[i].defaultbvalue
    else if defaults[i]._type = tString then
      PString(defaults[i].location)^ := defaults[i].defaultsvalue;

  // check for a custom default file
  i := M_CheckParm('-config');
  if (i > 0) and (i < myargc - 1) then
  begin
    defaultfile := myargv[i + 1];
    printf(' default file: %s' + #13#10, [defaultfile]);
  end
  else
    defaultfile := basedefault;

  s := TStringList.Create;
  try
    // read the file in, overriding any set defaults
    if fexists(defaultfile) then
      s.LoadFromFile(defaultfile);

    for i := 0 to s.Count - 1 do
    begin
      idx := -1;
      n := s.Names[i];
      for j := 0 to NUMDEFAULTS - 1 do
        if defaults[j].name = n then
        begin
          idx := j;
          break;
        end;

      if idx > -1 then
      begin
        if defaults[idx]._type = tInteger then
          PInteger(defaults[idx].location)^ := atoi(s.Values[n])
        else if defaults[idx]._type = tBoolean then
          PBoolean(defaults[idx].location)^ := boolval(atoi(s.Values[n]))
        else if defaults[idx]._type = tString then
          PString(defaults[idx].location)^ := s.Values[n];
      end;
    end;

  finally
    s.Free;
  end;
end;


type
  TargaHeader = record
    id_length, colormap_type, image_type: byte;
    colormap_index, colormap_length: word;
    colormap_size: byte;
    x_origin, y_origin, width, height: word;
    pixel_size, attributes: byte;
  end;

//
// M_ScreenShot
//
procedure M_ScreenShot;
var
  buffer: PByteArray;
  tganame: string;
  i: integer;
  buf: array[0..MAXWIDTH - 1] of LongWord;
  src: PByteArray;
begin
  tganame := 'DOOM00.tga';
//
// find a file name to save it to
//

  i := 0;
  while i <= 99 do
  begin
    tganame[5] := Chr((i div 10) + Ord('0'));
    tganame[6] := Chr((i mod 10) + Ord('0'));
    if not fexists(tganame) then
      break;  // file doesn't exist
    inc(i);
  end;
  if i = 100 then
  begin
    players[consoleplayer]._message := 'Couldn''t create a screenshot';
    exit;
  end;

  buffer := malloc(SCREENWIDTH * SCREENHEIGHT * 4 + 18);
  memset(buffer, 0, 18);
  buffer[2] := 2;    // uncompressed type
  buffer[12] := SCREENWIDTH and 255;
  buffer[13] := SCREENWIDTH div 256;
  buffer[14] := SCREENHEIGHT and 255;
  buffer[15] := SCREENHEIGHT div 256;
  buffer[16] := 32;  // pixel size

  src := @buffer[18];

  I_ReadScreen32bit(src);

  for i := 0 to (SCREENHEIGHT div 2) - 1 do
  begin
    memcpy(@buf, @src[i * SCREENWIDTH * 4], SCREENWIDTH * 4);
    memcpy(@src[i * SCREENWIDTH * 4], @src[(SCREENHEIGHT - i) * SCREENWIDTH * 4], SCREENWIDTH * 4);
    memcpy(@src[(SCREENHEIGHT - i) * SCREENWIDTH * 4], @buf, SCREENWIDTH * 4);
  end;

  if not M_WriteFile(tganame, buffer, SCREENWIDTH * SCREENHEIGHT * 4 + 18) then
    players[consoleplayer]._message := 'Couldn''t create a screenshot'
  else
    players[consoleplayer]._message := 'screen shot';

  freemem(buffer);

end;


end.

